function [model, lambda, gamma, history] = train_itr(y, K, mbr, vskl_options)
%function [model, lambda, gamma, history] = train_itr(y,K,mbr,vskl_options)
%Input:
% y - column vector of labels of training datapoints
% K - K(:,:,i) is the i^th training kernel (training examples X training examples). EACH KERNEL ASSUMED TO BE PD and UNIT-TRACE NORMALIZED.
% mbr - mbr(i)=j if i^th kernel belongs to j^th descriptor
% vskl_options - This is a structure containing the following parameters
%     C - SVM regularization parameter
%     q - paramter for the norm in reg. q can vary from 1 (L2-norm) to \infty (L-\infty norm)
%     sc - stopping criteria (sc=1 then rel.err in obj, sc=2 then variation in lambda)
%     svm_type - 
%     	svm_type = 1 (libsvm : Needs libsvm to be installed in the path)
%	svm_type = 2 (Simple MKL SVM solver : Needs Simple MKL to be installed)
%Output:
% model - libsvm model with optimal kernel
% lambda, gamma - optimal kernel weights
% history - a structure which contains the trace of the run, including the following
%          objlist - the list of obj values at the end of each outer loop iteration
%          inner_count - Total number of inner itertions
%          outer_count - Total number of Outer Iterations
%          converged - a boolean variable to denote if the algotithm converged within the max number of iterations
%          total_time - time taken to run each inner iteration       
%
% Authors : Saketha Nath.J - saketh@cse.iitb.ac.in, Raman - ramans@csa.iisc.ernet.in

%    Copyright (C) 2009 Saketha Nath.J
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.

objlist = [];

ITERMAX = vskl_options.itermax;
myEpsO = vskl_options.myEps0;
myEpsL = vskl_options.myEpsL;
C = vskl_options.C;
q = vskl_options.q;
sc = vskl_options.sc;
svm_type = vskl_options.svm_type;

m=size(y,1);
mpos = sum(y==1);
sn = size(K,3);
desc = unique(mbr);
n = length(desc);
if(any(desc~=(1:n)') || any(mbr~=sort(mbr)))
    fprintf('Format of mbr wrong!\n');
end
nvec=zeros(n,1);
for i=1:sn
	nvec(mbr(i)) = nvec(mbr(i)) + 1;
end
nbar = sqrt(log(max(nvec)));
svm_opt=sprintf('-q -t 4 -c %f -w1 %f -w-1 %f', C, (m-mpos)/m, mpos/m);

%Initialization of gamma
ITERMAX_OUT = ITERMAX;
if(q==1)
qs=inf;
gamma=ones(n,1);
ITERMAX_IN=1;
% ITERMAX_OUT=ITERMAX;
else
ITERMAX_IN=ITERMAX;
if(q==inf)
qs = 1;
else
qs = q/(q-1);
end
gamma=n^(-1/qs)*ones(n,1);
end

for i=1:sn
	lambda(i,1) = 1/nvec(mbr(i));
end
obj_old_out = eps;
obj_old_in = eps;

% options for the alternate SVm solver
lambdareg = 0;
kernel = 'numerical';
verbosesvm = 0;
span = 1;
alphainit = [];

inner_counts = [];
inner_count = 0;
outer_count = 0;
time_total = [];
history.converged = 1;

for iter_out=1:ITERMAX_OUT
    time_start = tic;
    if(iter_out == ITERMAX_OUT)
        history.converged = 0;
    end
    outer_count = outer_count + 1;
    inner_count = 0;
	for iter_in=1:ITERMAX_IN
        inner_count = inner_count + 1;
        K_mod = zeros(m,m);
        for i=1:sn
            K_mod = K_mod + lambda(i)*K(:,:,i)/gamma(mbr(i));
        end
	if(svm_type == 2)
            kerneloption.matrix = K_mod;
            [xsup,Alpsup,w0,pos,aux,aux,obj] = svmclass([], y, C, lambdareg,kernel,kerneloption,verbosesvm,span,alphainit);
            model.sv_coef = Alpsup;
            model.SVs = pos;
            model.rho = w0;
            model.totalSV = length(pos);
            model.nr_class = 2;       
	else
	    model = svmtrain(y,[(1:m)' K_mod],svm_opt);
        end
		obj = sum(abs(model.sv_coef))-0.5*model.sv_coef'*K_mod(model.SVs,model.SVs)*model.sv_coef;
%		fprintf('Obj at out=%d in=%d = %f\n',iter_out,iter_in,obj);
        if(obj < obj_old_in)
            if(iter_in > 1)
                fprintf('inner error\n');
            end
        end

        if(q ~= 1)
            dvec = zeros(n,1);
            for i=1:sn
                dvec(mbr(i)) = dvec(mbr(i)) + 0.5*model.sv_coef'*lambda(i)*K(model.SVs,model.SVs,i)*model.sv_coef;
            end
            gamma_new = dvec.^(qs/(1+qs));
            sumgamma_new = sum(gamma_new);
            gamma_new = (gamma_new./sumgamma_new).^(1/qs);
            
            if(sc==1)
                if(abs((obj-obj_old_in)/obj_old_in)<myEpsO)
                    obj_old_in = obj;
                    gamma = gamma_new;
                    break;
                end
            elseif(sc == 2)
                if(max(abs(gamma-gamma_new))<myEpsO)
                    obj_old_in = obj;
                    gamma = gamma_new;
                    break;
                end
            end
            gamma = gamma_new;
        end
		obj_old_in = obj;
    end
    inner_counts = [inner_counts; inner_count];
    if(obj > obj_old_out)
        if(iter_out > 1)
            fprintf('outer error\n');
        end
    end
    objlist = [objlist obj_old_in];
    for i=1:sn
        dvec(i,1) = -0.5*model.sv_coef'*K(model.SVs,model.SVs,i)*model.sv_coef/gamma(mbr(i));
    end
    st_size = nbar/(n*norm(dvec,inf)*sqrt(iter_in));
    dvec = n*st_size*dvec - log(lambda/n) - 1;
    lambda_new = exp(-dvec);
    lambda_new(find(lambda_new < myEpsL)) = 0;
    normsum=zeros(n,1);
    for i=1:sn
        normsum(mbr(i)) = normsum(mbr(i)) + lambda_new(i);
    end
    for i=1:sn
        lambda_new(i) = lambda_new(i)/normsum(mbr(i));
            if(abs(lambda_new(i)-lambda(i))>myEpsL)
            K_mod = K_mod + (lambda_new(i)-lambda(i))*K(:,:,i)/gamma(mbr(i));
        end
    end

    time_total = [time_total toc(time_start)];
    if(sc==1)
        if(abs((obj-obj_old_out)/obj_old_out)<myEpsO)
            obj_old_out = obj;
                lambda = lambda_new;
            break;
        end
    elseif(sc == 2)
        if(max(abs(lambda-lambda_new))<myEpsO)
            obj_old_out = obj;
                lambda = lambda_new;
            break;
        end
    end
    lambda = lambda_new;
    obj_old_out = obj;
end
history.total_time = time_total;
history.objlist = objlist;
history.outer_count = outer_count;
history.inner_counts = inner_counts;
return
